import logging
import socket as _socket
import ssl as _ssl
import time
import types
from _typeshed import Incomplete, Unused
from collections.abc import Callable
from typing import Any, TypeVar
from typing_extensions import TypeAlias

from .matcher import MQTTMatcher as MQTTMatcher
from .properties import Properties as Properties
from .reasoncodes import ReasonCodes as ReasonCodes
from .subscribeoptions import SubscribeOptions as SubscribeOptions

ssl: types.ModuleType | None
socks: types.ModuleType | None
time_func = time.monotonic
HAVE_DNS: bool
EAGAIN: int | Incomplete
MQTTv31: int
MQTTv311: int
MQTTv5: int
unicode = str
basestring = str
CONNECT: int
CONNACK: int
PUBLISH: int
PUBACK: int
PUBREC: int
PUBREL: int
PUBCOMP: int
SUBSCRIBE: int
SUBACK: int
UNSUBSCRIBE: int
UNSUBACK: int
PINGREQ: int
PINGRESP: int
DISCONNECT: int
AUTH: int
MQTT_LOG_INFO: int
MQTT_LOG_NOTICE: int
MQTT_LOG_WARNING: int
MQTT_LOG_ERR: int
MQTT_LOG_DEBUG: int
LOGGING_LEVEL: dict[int, int]
CONNACK_ACCEPTED: int
CONNACK_REFUSED_PROTOCOL_VERSION: int
CONNACK_REFUSED_IDENTIFIER_REJECTED: int
CONNACK_REFUSED_SERVER_UNAVAILABLE: int
CONNACK_REFUSED_BAD_USERNAME_PASSWORD: int
CONNACK_REFUSED_NOT_AUTHORIZED: int
mqtt_cs_new: int
mqtt_cs_connected: int
mqtt_cs_disconnecting: int
mqtt_cs_connect_async: int
mqtt_ms_invalid: int
mqtt_ms_publish: int
mqtt_ms_wait_for_puback: int
mqtt_ms_wait_for_pubrec: int
mqtt_ms_resend_pubrel: int
mqtt_ms_wait_for_pubrel: int
mqtt_ms_resend_pubcomp: int
mqtt_ms_wait_for_pubcomp: int
mqtt_ms_send_pubrec: int
mqtt_ms_queued: int
MQTT_ERR_AGAIN: int
MQTT_ERR_SUCCESS: int
MQTT_ERR_NOMEM: int
MQTT_ERR_PROTOCOL: int
MQTT_ERR_INVAL: int
MQTT_ERR_NO_CONN: int
MQTT_ERR_CONN_REFUSED: int
MQTT_ERR_NOT_FOUND: int
MQTT_ERR_CONN_LOST: int
MQTT_ERR_TLS: int
MQTT_ERR_PAYLOAD_SIZE: int
MQTT_ERR_NOT_SUPPORTED: int
MQTT_ERR_AUTH: int
MQTT_ERR_ACL_DENIED: int
MQTT_ERR_UNKNOWN: int
MQTT_ERR_ERRNO: int
MQTT_ERR_QUEUE_SIZE: int
MQTT_ERR_KEEPALIVE: int
MQTT_CLIENT: int
MQTT_BRIDGE: int
MQTT_CLEAN_START_FIRST_ONLY: int
sockpair_data: bytes
_UserData: TypeAlias = Any
_Socket: TypeAlias = _socket.socket | _ssl.SSLSocket | Incomplete
_Payload: TypeAlias = str | bytes | bytearray | float
_ExtraHeader: TypeAlias = dict[str, str] | Callable[[dict[str, str]], dict[str, str]]
_OnLog: TypeAlias = Callable[[Client, _UserData, int, str], object]
_OnConnect: TypeAlias = Callable[[Client, _UserData, dict[str, int], int], object]
_OnConnectV5: TypeAlias = Callable[[Client, _UserData, dict[str, int], ReasonCodes, Properties | None], object]
_TOnConnect = TypeVar("_TOnConnect", _OnConnect, _OnConnectV5)
_OnConnectFail: TypeAlias = Callable[[Client, _UserData], object]
_OnSubscribe: TypeAlias = Callable[[Client, _UserData, int, tuple[int]], object]
_OnSubscribeV5: TypeAlias = Callable[[Client, _UserData, int, list[ReasonCodes], Properties], object]
_TOnSubscribe = TypeVar("_TOnSubscribe", _OnSubscribe, _OnSubscribeV5)
_OnMessage: TypeAlias = Callable[[Client, _UserData, MQTTMessage], object]
_OnPublish: TypeAlias = Callable[[Client, _UserData, int], object]
_OnUnsubscribe: TypeAlias = Callable[[Client, _UserData, int], object]
_OnUnsubscribeV5: TypeAlias = Callable[[Client, _UserData, int, Properties, list[ReasonCodes] | ReasonCodes], object]
_TOnUnsubscribe = TypeVar("_TOnUnsubscribe", _OnUnsubscribe, _OnUnsubscribeV5)
_OnDisconnect: TypeAlias = Callable[[Client, _UserData, int], object]
_OnDisconnectV5: TypeAlias = Callable[[Client, _UserData, ReasonCodes | None, Properties | None], object]
_TOnDisconnect = TypeVar("_TOnDisconnect", _OnDisconnect, _OnDisconnectV5)
_OnSocket: TypeAlias = Callable[[Client, _UserData, _Socket | WebsocketWrapper | None], object]

class WebsocketConnectionError(ValueError): ...

def error_string(mqtt_errno: int) -> str: ...
def connack_string(connack_code: int) -> str: ...
def base62(num: int, base: str = ..., padding: int = 1) -> str: ...
def topic_matches_sub(sub: str, topic: str) -> bool: ...

class MQTTMessageInfo:
    mid: int
    rc: int
    def __init__(self, mid: int) -> None: ...
    def __iter__(self) -> MQTTMessageInfo: ...
    def __next__(self) -> int: ...
    def next(self) -> int: ...
    def __getitem__(self, index: int) -> int: ...
    def wait_for_publish(self, timeout: float | None = None) -> None: ...
    def is_published(self) -> bool: ...

class MQTTMessage:
    timestamp: int
    state: int
    dup: bool
    mid: int
    payload: bytes | bytearray
    qos: int
    retain: bool
    info: MQTTMessageInfo
    properties: Properties | None
    def __init__(self, mid: int = 0, topic: bytes | bytearray = b"") -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    @property
    def topic(self) -> str: ...
    @topic.setter
    def topic(self, value: bytes | bytearray) -> None: ...

class Client:
    suppress_exceptions: bool
    def __init__(
        self,
        client_id: str | None = "",
        clean_session: bool | None = None,
        userdata: _UserData | None = None,
        protocol: int = 4,
        transport: str = "tcp",
        reconnect_on_failure: bool = True,
    ) -> None: ...
    def __del__(self) -> None: ...
    def reinitialise(self, client_id: str = "", clean_session: bool = True, userdata: _UserData | None = None) -> None: ...
    def ws_set_options(self, path: str = "/mqtt", headers: _ExtraHeader | None = None) -> None: ...
    def tls_set_context(self, context: _ssl.SSLContext | None = None) -> None: ...
    def tls_set(
        self,
        ca_certs: str | None = None,
        certfile: str | None = None,
        keyfile: str | None = None,
        cert_reqs: _ssl.VerifyMode | None = None,
        tls_version: _ssl._SSLMethod | None = None,
        ciphers: str | None = None,
        keyfile_password: _ssl._PasswordType | None = None,
    ) -> None: ...
    def tls_insecure_set(self, value: bool) -> None: ...
    def proxy_set(self, **proxy_args: Any) -> None: ...
    def enable_logger(self, logger: logging.Logger | None = None) -> None: ...
    def disable_logger(self) -> None: ...
    def connect(
        self,
        host: str,
        port: int = 1883,
        keepalive: int = 60,
        bind_address: str = "",
        bind_port: int = 0,
        clean_start: int = 3,
        properties: Properties | None = None,
    ) -> int: ...
    def connect_srv(
        self,
        domain: str | None = None,
        keepalive: int = 60,
        bind_address: str = "",
        clean_start: int = 3,
        properties: Properties | None = None,
    ) -> int: ...
    def connect_async(
        self,
        host: str,
        port: int = 1883,
        keepalive: int = 60,
        bind_address: str = "",
        bind_port: int = 0,
        clean_start: int = 3,
        properties: Properties | None = None,
    ) -> None: ...
    def reconnect_delay_set(self, min_delay: int = 1, max_delay: int = 120) -> None: ...
    def reconnect(self) -> int: ...
    def loop(self, timeout: float = 1.0, max_packets: int = 1) -> int: ...
    def publish(
        self,
        topic: str,
        payload: _Payload | None = None,
        qos: int = 0,
        retain: bool = False,
        properties: Properties | None = None,
    ) -> MQTTMessageInfo: ...
    def username_pw_set(self, username: str, password: str | bytes | bytearray | None = None) -> None: ...
    def enable_bridge_mode(self) -> None: ...
    def is_connected(self) -> bool: ...
    def disconnect(self, reasoncode: ReasonCodes | None = None, properties: Properties | None = None) -> int: ...
    def subscribe(
        self,
        topic: str | tuple[str, SubscribeOptions] | list[tuple[str, SubscribeOptions]] | list[tuple[str, int]],
        qos: int = 0,
        options: SubscribeOptions | None = None,
        properties: Properties | None = None,
    ) -> tuple[int, int]: ...
    def unsubscribe(self, topic: str | list[str], properties: Properties | None = None) -> tuple[int, int]: ...
    def loop_read(self, max_packets: int = 1) -> int: ...
    def loop_write(self, max_packets: int = 1) -> int: ...
    def want_write(self) -> bool: ...
    def loop_misc(self) -> int: ...
    def max_inflight_messages_set(self, inflight: int) -> None: ...
    def max_queued_messages_set(self, queue_size: int) -> Client: ...
    def message_retry_set(self, retry: Unused) -> None: ...
    def user_data_set(self, userdata: _UserData) -> None: ...
    def will_set(
        self,
        topic: str,
        payload: _Payload | None = None,
        qos: int = 0,
        retain: bool = False,
        properties: Properties | None = None,
    ) -> None: ...
    def will_clear(self) -> None: ...
    def socket(self) -> _Socket | WebsocketWrapper: ...
    def loop_forever(self, timeout: float = 1.0, max_packets: int = 1, retry_first_connection: bool = False) -> int: ...
    def loop_start(self) -> int | None: ...
    def loop_stop(self, force: bool = False) -> int | None: ...
    @property
    def on_log(self) -> _OnLog | None: ...
    @on_log.setter
    def on_log(self, func: _OnLog | None) -> None: ...
    def log_callback(self) -> Callable[[_OnLog], _OnLog]: ...
    @property
    def on_connect(self) -> _OnConnect | _OnConnectV5 | None: ...
    @on_connect.setter
    def on_connect(self, func: _OnConnect | _OnConnectV5 | None) -> None: ...
    def connect_callback(self) -> Callable[[_TOnConnect], _TOnConnect]: ...
    @property
    def on_connect_fail(self) -> _OnConnectFail | None: ...
    @on_connect_fail.setter
    def on_connect_fail(self, func: _OnConnectFail | None) -> None: ...
    def connect_fail_callback(self) -> Callable[[_OnConnectFail], _OnConnectFail]: ...
    @property
    def on_subscribe(self) -> _OnSubscribe | _OnSubscribeV5 | None: ...
    @on_subscribe.setter
    def on_subscribe(self, func: _OnSubscribe | _OnSubscribeV5 | None) -> None: ...
    def subscribe_callback(self) -> Callable[[_TOnSubscribe], _TOnSubscribe]: ...
    @property
    def on_message(self) -> _OnMessage | None: ...
    @on_message.setter
    def on_message(self, func: _OnMessage | None) -> None: ...
    def message_callback(self) -> Callable[[_OnMessage], _OnMessage]: ...
    @property
    def on_publish(self) -> _OnPublish | None: ...
    @on_publish.setter
    def on_publish(self, func: _OnPublish | None) -> None: ...
    def publish_callback(self) -> Callable[[_OnPublish], _OnPublish]: ...
    @property
    def on_unsubscribe(self) -> _OnUnsubscribe | _OnUnsubscribeV5 | None: ...
    @on_unsubscribe.setter
    def on_unsubscribe(self, func: _OnUnsubscribe | _OnUnsubscribeV5 | None) -> None: ...
    def unsubscribe_callback(self) -> Callable[[_TOnUnsubscribe], _TOnUnsubscribe]: ...
    @property
    def on_disconnect(self) -> _OnDisconnect | _OnDisconnectV5 | None: ...
    @on_disconnect.setter
    def on_disconnect(self, func: _OnDisconnect | _OnDisconnectV5 | None) -> None: ...
    def disconnect_callback(self) -> Callable[[_TOnDisconnect], _TOnDisconnect]: ...
    @property
    def on_socket_open(self) -> _OnSocket | None: ...
    @on_socket_open.setter
    def on_socket_open(self, func: _OnSocket | None) -> None: ...
    def socket_open_callback(self) -> Callable[[_OnSocket], _OnSocket]: ...
    @property
    def on_socket_close(self) -> _OnSocket | None: ...
    @on_socket_close.setter
    def on_socket_close(self, func: _OnSocket | None) -> None: ...
    def socket_close_callback(self) -> Callable[[_OnSocket], _OnSocket]: ...
    @property
    def on_socket_register_write(self) -> _OnSocket | None: ...
    @on_socket_register_write.setter
    def on_socket_register_write(self, func: _OnSocket | None) -> None: ...
    def socket_register_write_callback(self) -> Callable[[_OnSocket], _OnSocket]: ...
    @property
    def on_socket_unregister_write(self) -> _OnSocket | None: ...
    @on_socket_unregister_write.setter
    def on_socket_unregister_write(self, func: _OnSocket | None) -> None: ...
    def socket_unregister_write_callback(self) -> Callable[[_OnSocket], _OnSocket]: ...
    def message_callback_add(self, sub: str, callback: _OnMessage) -> None: ...
    def topic_callback(self, sub: str) -> Callable[[_OnMessage], _OnMessage]: ...
    def message_callback_remove(self, sub: str) -> None: ...

class WebsocketWrapper:
    OPCODE_CONTINUATION: int
    OPCODE_TEXT: int
    OPCODE_BINARY: int
    OPCODE_CONNCLOSE: int
    OPCODE_PING: int
    OPCODE_PONG: int
    connected: bool
    def __init__(
        self, socket: _Socket, host: str, port: int, is_ssl: bool, path: str, extra_headers: _ExtraHeader | None
    ) -> None: ...
    def __del__(self) -> None: ...
    def recv(self, length: int) -> bytes | bytearray | None: ...
    def read(self, length: int) -> bytes | bytearray | None: ...
    def send(self, data: bytes | bytearray) -> int: ...
    def write(self, data: bytes | bytearray) -> int: ...
    def close(self) -> None: ...
    def fileno(self) -> int: ...
    def pending(self) -> int: ...
    def setblocking(self, flag: bool) -> None: ...



import enum
import sys
from typing import AnyStr

__all__ = ["compile", "main", "PyCompileError", "PycInvalidationMode"]

class PyCompileError(Exception):
    exc_type_name: str
    exc_value: BaseException
    file: str
    msg: str
    def __init__(self, exc_type: type[BaseException], exc_value: BaseException, file: str, msg: str = "") -> None: ...

class PycInvalidationMode(enum.Enum):
    TIMESTAMP: int
    CHECKED_HASH: int
    UNCHECKED_HASH: int

def _get_default_invalidation_mode() -> PycInvalidationMode: ...
def compile(
    file: AnyStr,
    cfile: AnyStr | None = None,
    dfile: AnyStr | None = None,
    doraise: bool = False,
    optimize: int = -1,
    invalidation_mode: PycInvalidationMode | None = None,
    quiet: int = 0,
) -> AnyStr | None: ...

if sys.version_info >= (3, 10):
    def main() -> None: ...

else:
    def main(args: list[str] | None = None) -> int: ...



from _typeshed import Incomplete

class PayPalAccountGateway:
    gateway: Incomplete
    config: Incomplete
    def __init__(self, gateway) -> None: ...
    def find(self, paypal_account_token): ...
    def delete(self, paypal_account_token): ...
    def update(self, paypal_account_token, params: Incomplete | None = None): ...



from _typeshed import Incomplete

text_type = str

class WebhookNotificationGateway:
    gateway: Incomplete
    config: Incomplete
    def __init__(self, gateway) -> None: ...
    def parse(self, signature, payload): ...
    def verify(self, challenge): ...



import logging

class MenuBorderStyle:
    @property
    def bottom_left_corner(self) -> str: ...
    @property
    def bottom_right_corner(self) -> str: ...
    @property
    def inner_horizontal(self) -> str: ...
    @property
    def inner_vertical(self) -> str: ...
    @property
    def intersection(self) -> str: ...
    @property
    def outer_horizontal(self) -> str: ...
    @property
    def outer_horizontal_inner_down(self) -> str: ...
    @property
    def outer_horizontal_inner_up(self) -> str: ...
    @property
    def outer_vertical(self) -> str: ...
    @property
    def outer_vertical_inner_left(self) -> str: ...
    @property
    def outer_vertical_inner_right(self) -> str: ...
    @property
    def top_left_corner(self) -> str: ...
    @property
    def top_right_corner(self) -> str: ...

class AsciiBorderStyle(MenuBorderStyle):
    @property
    def bottom_left_corner(self) -> str: ...
    @property
    def bottom_right_corner(self) -> str: ...
    @property
    def inner_horizontal(self) -> str: ...
    @property
    def inner_vertical(self) -> str: ...
    @property
    def intersection(self) -> str: ...
    @property
    def outer_horizontal(self) -> str: ...
    @property
    def outer_horizontal_inner_down(self) -> str: ...
    @property
    def outer_horizontal_inner_up(self) -> str: ...
    @property
    def outer_vertical(self) -> str: ...
    @property
    def outer_vertical_inner_left(self) -> str: ...
    @property
    def outer_vertical_inner_right(self) -> str: ...
    @property
    def top_left_corner(self) -> str: ...
    @property
    def top_right_corner(self) -> str: ...

class LightBorderStyle(MenuBorderStyle):
    @property
    def bottom_left_corner(self) -> str: ...
    @property
    def bottom_right_corner(self) -> str: ...
    @property
    def inner_horizontal(self) -> str: ...
    @property
    def inner_vertical(self) -> str: ...
    @property
    def intersection(self) -> str: ...
    @property
    def outer_horizontal(self) -> str: ...
    @property
    def outer_horizontal_inner_down(self) -> str: ...
    @property
    def outer_horizontal_inner_up(self) -> str: ...
    @property
    def outer_vertical(self) -> str: ...
    @property
    def outer_vertical_inner_left(self) -> str: ...
    @property
    def outer_vertical_inner_right(self) -> str: ...
    @property
    def top_left_corner(self) -> str: ...
    @property
    def top_right_corner(self) -> str: ...

class HeavyBorderStyle(MenuBorderStyle):
    @property
    def bottom_left_corner(self) -> str: ...
    @property
    def bottom_right_corner(self) -> str: ...
    @property
    def inner_horizontal(self) -> str: ...
    @property
    def inner_vertical(self) -> str: ...
    @property
    def intersection(self) -> str: ...
    @property
    def outer_horizontal(self) -> str: ...
    @property
    def outer_horizontal_inner_down(self) -> str: ...
    @property
    def outer_horizontal_inner_up(self) -> str: ...
    @property
    def outer_vertical(self) -> str: ...
    @property
    def outer_vertical_inner_left(self) -> str: ...
    @property
    def outer_vertical_inner_right(self) -> str: ...
    @property
    def top_left_corner(self) -> str: ...
    @property
    def top_right_corner(self) -> str: ...

class HeavyOuterLightInnerBorderStyle(HeavyBorderStyle):
    @property
    def inner_horizontal(self) -> str: ...
    @property
    def inner_vertical(self) -> str: ...
    @property
    def intersection(self) -> str: ...
    @property
    def outer_horizontal_inner_down(self) -> str: ...
    @property
    def outer_horizontal_inner_up(self) -> str: ...
    @property
    def outer_vertical_inner_left(self) -> str: ...
    @property
    def outer_vertical_inner_right(self) -> str: ...

class DoubleLineBorderStyle(MenuBorderStyle):
    @property
    def bottom_left_corner(self) -> str: ...
    @property
    def bottom_right_corner(self) -> str: ...
    @property
    def inner_horizontal(self) -> str: ...
    @property
    def inner_vertical(self) -> str: ...
    @property
    def intersection(self) -> str: ...
    @property
    def outer_horizontal(self) -> str: ...
    @property
    def outer_horizontal_inner_down(self) -> str: ...
    @property
    def outer_horizontal_inner_up(self) -> str: ...
    @property
    def outer_vertical(self) -> str: ...
    @property
    def outer_vertical_inner_left(self) -> str: ...
    @property
    def outer_vertical_inner_right(self) -> str: ...
    @property
    def top_left_corner(self) -> str: ...
    @property
    def top_right_corner(self) -> str: ...

class DoubleLineOuterLightInnerBorderStyle(DoubleLineBorderStyle):
    @property
    def inner_horizontal(self) -> str: ...
    @property
    def inner_vertical(self) -> str: ...
    @property
    def intersection(self) -> str: ...
    @property
    def outer_horizontal_inner_down(self) -> str: ...
    @property
    def outer_horizontal_inner_up(self) -> str: ...
    @property
    def outer_vertical_inner_left(self) -> str: ...
    @property
    def outer_vertical_inner_right(self) -> str: ...

class MenuBorderStyleType:
    ASCII_BORDER: int
    LIGHT_BORDER: int
    HEAVY_BORDER: int
    DOUBLE_LINE_BORDER: int
    HEAVY_OUTER_LIGHT_INNER_BORDER: int
    DOUBLE_LINE_OUTER_LIGHT_INNER_BORDER: int

class MenuBorderStyleFactory:
    logger: logging.Logger
    def __init__(self) -> None: ...
    def create_border(self, border_style_type: MenuBorderStyleType) -> MenuBorderStyle: ...
    def create_ascii_border(self) -> AsciiBorderStyle: ...
    def create_light_border(self) -> LightBorderStyle: ...
    def create_heavy_border(self) -> HeavyBorderStyle: ...
    def create_heavy_outer_light_inner_border(self) -> HeavyOuterLightInnerBorderStyle: ...
    def create_doubleline_border(self) -> DoubleLineBorderStyle: ...
    def create_doubleline_outer_light_inner_border(self) -> DoubleLineOuterLightInnerBorderStyle: ...
    @staticmethod
    def is_win_python35_or_earlier() -> bool: ...



from _typeshed import Incomplete
from pathlib import Path
from re import Pattern
from typing import ClassVar

from docutils.parsers.rst import Directive
from docutils.parsers.rst.states import SpecializedBody

__docformat__: str

class Include(Directive):
    standard_include_path: Path

class Raw(Directive): ...
class Replace(Directive): ...

class Unicode(Directive):
    comment_pattern: Pattern[str]

class Class(Directive): ...

class Role(Directive):
    argument_pattern: Pattern[str]

class DefaultRole(Directive): ...
class Title(Directive): ...

# SpecializedBody has not yet been stubbed
class MetaBody(SpecializedBody):  # pyright: ignore[reportUntypedBaseClass]
    def __getattr__(self, name: str) -> Incomplete: ...

class Meta(Directive):
    SMkwargs: ClassVar[dict[str, tuple[MetaBody]]]

class Date(Directive): ...
class TestDirective(Directive): ...



from _typeshed import Incomplete
from enum import IntEnum
from typing import ClassVar, Literal

from .ImageFile import ImageFile, PyDecoder, PyEncoder

class Format(IntEnum):
    JPEG: int

class Encoding(IntEnum):
    UNCOMPRESSED: int
    DXT: int
    UNCOMPRESSED_RAW_BGRA: int

class AlphaEncoding(IntEnum):
    DXT1: int
    DXT3: int
    DXT5: int

def unpack_565(i): ...
def decode_dxt1(data, alpha: bool = False): ...
def decode_dxt3(data): ...
def decode_dxt5(data): ...

class BLPFormatError(NotImplementedError): ...

class BlpImageFile(ImageFile):
    format: ClassVar[Literal["BLP"]]
    format_description: ClassVar[str]

class _BLPBaseDecoder(PyDecoder):
    magic: Incomplete
    def decode(self, buffer): ...

class BLP1Decoder(_BLPBaseDecoder): ...
class BLP2Decoder(_BLPBaseDecoder): ...

class BLPEncoder(PyEncoder):
    def encode(self, bufsize): ...



from _typeshed import Incomplete

from oauthlib.oauth2 import Client
from requests.auth import AuthBase

class OAuth2(AuthBase):
    def __init__(
        self, client_id: Incomplete | None = None, client: Client | None = None, token: Incomplete | None = None
    ) -> None: ...



from _typeshed import Incomplete
from logging import Logger
from typing import TypedDict
from typing_extensions import TypeAlias

import requests
from oauthlib.oauth1 import Client

from . import OAuth1

# should be dict[str, str] but could look different
_ParsedToken: TypeAlias = dict[str, Incomplete]

class _TokenDict(TypedDict, total=False):
    oauth_token: Incomplete  # oauthlib.oauth1.Client.resource_owner_key
    oauth_token_secret: Incomplete  # oauthlib.oauth1.Client.resource_token_secret
    oauth_verifier: Incomplete  # oauthlib.oauth1.Client.oauth_verifier

log: Logger

def urldecode(body): ...

class TokenRequestDenied(ValueError):
    response: requests.Response
    def __init__(self, message: str, response: requests.Response) -> None: ...
    @property
    def status_code(self) -> int: ...

class TokenMissing(ValueError):
    response: requests.Response
    def __init__(self, message: str, response: requests.Response) -> None: ...

class VerifierMissing(ValueError): ...

class OAuth1Session(requests.Session):
    auth: OAuth1
    def __init__(
        self,
        client_key,
        client_secret: Incomplete | None = None,
        resource_owner_key: Incomplete | None = None,
        resource_owner_secret: Incomplete | None = None,
        callback_uri: Incomplete | None = None,
        signature_method="HMAC-SHA1",
        signature_type="AUTH_HEADER",
        rsa_key: Incomplete | None = None,
        verifier: Incomplete | None = None,
        client_class: type[Client] | None = None,
        force_include_body: bool = False,
        *,
        encoding: str = "utf-8",
        nonce: Incomplete | None = None,
        timestamp: Incomplete | None = None,
    ) -> None: ...
    @property
    def token(self) -> _TokenDict: ...
    @token.setter
    def token(self, value: _TokenDict) -> None: ...
    @property
    def authorized(self) -> bool: ...
    def authorization_url(self, url: str, request_token: Incomplete | None = None, **kwargs) -> str: ...
    def fetch_request_token(self, url: str, realm: Incomplete | None = None, **request_kwargs) -> _ParsedToken: ...
    def fetch_access_token(self, url: str, verifier: Incomplete | None = None, **request_kwargs) -> _ParsedToken: ...
    def parse_authorization_response(self, url: str) -> _ParsedToken: ...
    def rebuild_auth(self, prepared_request: requests.PreparedRequest, response: requests.Response) -> None: ...



from collections.abc import Callable, Generator, Iterable, Iterator, Sequence
from re import Match, Pattern
from typing import Any, ClassVar, Generic, SupportsIndex, TypeVar, overload
from typing_extensions import Self, TypeAlias

_T = TypeVar("_T")
_Context = TypeVar("_Context")
_TransitionResult: TypeAlias = tuple[_Context, str | None, list[str]]
_TransitionMethod: TypeAlias = Callable[[Match[str], _Context, str], _TransitionResult[_Context]]
_Observer: TypeAlias = Callable[[StateMachine[_Context]], None]

class StateMachine(Generic[_Context]):
    input_lines: StringList | None
    input_offset: int
    line: str | None
    line_offset: int
    debug: bool
    initial_state: str
    current_state: str
    states: dict[str, State[_Context]]
    observers: list[_Observer[_Context]]
    def __init__(self, state_classes: Iterable[type[State[_Context]]], initial_state: str, debug: bool = False) -> None: ...
    def unlink(self) -> None: ...
    def run(
        self,
        input_lines: Sequence[str] | StringList,
        input_offset: int = 0,
        context: _Context | None = None,
        input_source: str | None = None,
        initial_state: str | None = None,
    ) -> list[str]: ...
    def get_state(self, next_state: str | None = None) -> State[_Context]: ...
    def next_line(self, n: int = 1) -> str: ...
    def is_next_line_blank(self) -> bool: ...
    def at_eof(self) -> bool: ...
    def at_bof(self) -> bool: ...
    def previous_line(self, n: int = 1) -> str | None: ...
    def goto_line(self, line_offset: int) -> str | None: ...
    def get_source(self, line_offset: int) -> str: ...
    def abs_line_offset(self) -> int: ...
    def abs_line_number(self) -> int: ...
    def get_source_and_line(self, lineno: int | None = None) -> tuple[str, int] | tuple[None, None]: ...
    def insert_input(self, input_lines: list[str] | StringList, source: str) -> None: ...
    def get_text_block(self, flush_left: bool = False) -> StringList: ...
    def check_line(
        self, context: _Context, state: State[_Context], transitions: list[str] | None = ...
    ) -> _TransitionResult[_Context]: ...
    def add_state(self, state_class: type[State[_Context]]) -> None: ...
    def add_states(self, state_classes: Iterable[type[State[_Context]]]) -> None: ...
    def runtime_init(self) -> None: ...
    def error(self) -> None: ...
    def attach_observer(self, observer: _Observer[_Context]) -> None: ...
    def detach_observer(self, observer: _Observer[_Context]) -> None: ...
    def notify_observers(self) -> None: ...

class State(Generic[_Context]):
    patterns: ClassVar[dict[str, str | Pattern[str]] | None]
    initial_transitions: ClassVar[Sequence[str] | Sequence[tuple[str, str]] | None]
    nested_sm: type[StateMachine[_Context]]
    nested_sm_kwargs: dict[str, Any]
    transition_order: list[str]
    transitions: dict[str, tuple[Pattern[str], Callable[[], None], str]]
    state_machine: StateMachine[_Context]
    debug: bool
    def __init__(self, state_machine: StateMachine[_Context], debug: bool = False) -> None: ...
    def runtime_init(self) -> None: ...
    def unlink(self) -> None: ...
    def add_initial_transitions(self) -> None: ...
    def add_transitions(self, names: Iterable[str], transitions) -> None: ...
    def add_transition(self, name: str, transition: tuple[Pattern[str], str, str]) -> None: ...
    def remove_transition(self, name: str) -> None: ...
    def make_transition(
        self, name: str, next_state: str | None = None
    ) -> tuple[Pattern[str], _TransitionMethod[_Context], str]: ...
    def make_transitions(
        self, name_list: list[str | tuple[str] | tuple[str, str]]
    ) -> tuple[list[str], dict[str, tuple[Pattern[str], _TransitionMethod[_Context], str]]]: ...
    def no_match(
        self, context: _Context, transitions: tuple[list[str], dict[str, tuple[Pattern[str], _TransitionMethod[_Context], str]]]
    ) -> _TransitionResult[_Context]: ...
    def bof(self, context: _Context) -> tuple[list[str], list[str]]: ...
    def eof(self, context: _Context) -> list[str]: ...
    def nop(self, match: Match[str], context: _Context, next_state: str) -> _TransitionResult[_Context]: ...

class StateMachineWS(StateMachine[_Context]):
    def get_indented(self, until_blank: bool = False, strip_indent: bool = True) -> tuple[StringList, int, int, bool]: ...
    def get_known_indented(
        self, indent: int, until_blank: bool = False, strip_indent: bool = True
    ) -> tuple[list[str], int, bool]: ...
    def get_first_known_indented(
        self, indent: int, until_blank: bool = False, strip_indent: bool = True, strip_top: bool = True
    ) -> tuple[list[str], int, int, bool]: ...

class StateWS(State[_Context]):
    indent_sm: type[StateMachine[_Context]] | None
    indent_sm_kwargs: dict[str, Any] | None
    known_indent_sm: type[StateMachine[_Context]] | None
    known_indent_sm_kwargs: dict[str, Any] | None
    ws_patterns: dict[str, Pattern[str]]
    ws_initial_transitions: Sequence[str]
    def __init__(self, state_machine: StateMachine[_Context], debug: bool = False) -> None: ...
    def add_initial_transitions(self) -> None: ...
    def blank(self, match: Match[str], context: _Context, next_state: str) -> _TransitionResult[_Context]: ...
    def indent(self, match: Match[str], context: _Context, next_state: str) -> _TransitionResult[_Context]: ...
    def known_indent(self, match: Match[str], context: _Context, next_state: str) -> _TransitionResult[_Context]: ...
    def first_known_indent(self, match: Match[str], context: _Context, next_state: str) -> _TransitionResult[_Context]: ...

class _SearchOverride:
    def match(self, pattern: Pattern[str]) -> Match[str]: ...

class SearchStateMachine(_SearchOverride, StateMachine[_Context]): ...
class SearchStateMachineWS(_SearchOverride, StateMachineWS[_Context]): ...

class ViewList(Generic[_T]):
    data: list[_T]
    items: list[tuple[str, int]]
    parent: Self
    parent_offset: int
    def __init__(
        self,
        initlist: Self | Sequence[_T] | None = None,
        source: str | None = None,
        items: list[tuple[str, int]] | None = None,
        parent: Self | None = None,
        parent_offset: int | None = None,
    ) -> None: ...
    def __lt__(self, other: Any) -> bool: ...
    def __le__(self, other: Any) -> bool: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __gt__(self, other: Any) -> bool: ...
    def __ge__(self, other: Any) -> bool: ...
    def __contains__(self, item: _T) -> bool: ...
    def __len__(self) -> int: ...
    @overload
    def __getitem__(self, i: slice) -> Self: ...
    @overload
    def __getitem__(self, i: SupportsIndex) -> _T: ...
    @overload
    def __setitem__(self, i: slice, item: Self) -> None: ...
    @overload
    def __setitem__(self, i: SupportsIndex, item: _T) -> None: ...
    def __delitem__(self, i: SupportsIndex) -> None: ...
    def __add__(self, other: Self) -> Self: ...
    def __radd__(self, other: Self) -> Self: ...
    def __iadd__(self, other: Self) -> Self: ...
    def __mul__(self, n: int) -> Self: ...
    __rmul__ = __mul__
    def __imul__(self, n: int) -> Self: ...
    def extend(self, other: Self) -> None: ...
    def append(self, item: _T, source: str | None = None, offset: int = 0) -> None: ...
    def insert(self, i: int, item: _T, source: str | None = None, offset: int = 0) -> None: ...
    def pop(self, i: int = -1) -> _T: ...
    def trim_start(self, n: int = 1) -> None: ...
    def trim_end(self, n: int = 1) -> None: ...
    def remove(self, item: _T) -> None: ...
    def count(self, item: _T) -> int: ...
    def index(self, item: _T) -> int: ...
    def reverse(self) -> None: ...
    def sort(self, *args: tuple[_T, tuple[str, int]]) -> None: ...
    def info(self, i: int) -> tuple[str, int | None]: ...
    def source(self, i: int) -> str: ...
    def offset(self, i: int) -> int: ...
    def disconnect(self) -> None: ...
    def xitems(self) -> Generator[tuple[str, int, str], None, None]: ...
    def pprint(self) -> None: ...

    # dummy atribute to indicate to mypy that ViewList is Iterable[str]
    def __iter__(self) -> Iterator[str]: ...

class StringList(ViewList[str]):
    def trim_left(self, length: int, start: int = 0, end: int = ...) -> None: ...
    def get_text_block(self, start: int, flush_left: bool = False) -> StringList: ...
    def get_indented(
        self,
        start: int = 0,
        until_blank: bool = False,
        strip_indent: bool = True,
        block_indent: int | None = None,
        first_indent: int | None = None,
    ) -> tuple[StringList, int, bool]: ...
    def get_2D_block(self, top: int, left: int, bottom: int, right: int, strip_indent: bool = True) -> StringList: ...
    def pad_double_width(self, pad_char: str) -> None: ...
    def replace(self, old: str, new: str) -> None: ...

class StateMachineError(Exception): ...
class UnknownStateError(StateMachineError): ...
class DuplicateStateError(StateMachineError): ...
class UnknownTransitionError(StateMachineError): ...
class DuplicateTransitionError(StateMachineError): ...
class TransitionPatternNotFound(StateMachineError): ...
class TransitionMethodNotFound(StateMachineError): ...
class UnexpectedIndentationError(StateMachineError): ...
class TransitionCorrection(Exception): ...
class StateCorrection(Exception): ...

def string2lines(
    astring: str, tab_width: int = 8, convert_whitespace: bool = False, whitespace: Pattern[str] = ...
) -> list[str]: ...
